@@ -31,6 +31,16 @@ param_errmsg = "parameter/artifact name must consist of alpha-numeric "\
 key_regex = re.compile("^[a-z0-9][-a-z0-9]*$")
 key_errmsg = "key must consist of lower case alpha-numeric characters or '-',"\
     "and must start with an alpha-numeric character (e.g. 'my-key')"
+input_parameter_pattern = re.compile(r"^{{inputs\.parameters\.(.*)}}$")
+input_artifact_pattern = re.compile(r"^{{inputs\.artifacts\.(.*)}}$")
+step_output_parameter_pattern = re.compile(
+    r"^{{steps\.(.*?)\.outputs\.parameters\.(.*?)}}$")
+step_output_artifact_pattern = re.compile(
+    r"^{{steps\.(.*?)\.outputs\.artifacts\.(.*?)}}$")
+task_output_parameter_pattern = re.compile(
+    r"^{{tasks\.(.*?)\.outputs\.parameters\.(.*?)}}$")
+task_output_artifact_pattern = re.compile(
+    r"^{{tasks\.(.*?)\.outputs\.artifacts\.(.*?)}}$")
 
 
 class S3Artifact(V1alpha1S3Artifact):
@@ -43,20 +53,23 @@ class S3Artifact(V1alpha1S3Artifact):
 
     def __init__(
             self,
+            key: str = None,
             path_list: Union[str, list] = None,
             urn: str = "",
+            debug_s3: bool = False,
             *args,
             **kwargs,
     ) -> None:
         config = Configuration()
         config.client_side_validation = False
-        super().__init__(local_vars_configuration=config, *args, **kwargs)
+        super().__init__(local_vars_configuration=config, key=key, *args,
+                         **kwargs)
         if urn:
             if global_config["lineage"]:
                 meta = global_config["lineage"].get_artifact_metadata(urn)
                 self.key = meta.uri
             else:
-                logging.warn("Lineage client not provided")
+                logging.warning("Lineage client not provided")
         assert isinstance(self.key, str)
         if not self.key.startswith(s3_config["prefix"]) and not any(
                 [self.key.startswith(p) for p in s3_config["extra_prefixes"]]):
@@ -67,6 +80,10 @@ class S3Artifact(V1alpha1S3Artifact):
         self.urn = urn
         self.slice = None
         self.parent = None
+        self.local_path = None
+        if global_config["mode"] == "debug" and not global_config["debug_s3"] \
+                and not debug_s3:
+            self.local_path = self.key
 
     def __getitem__(self, key):
         art = copy(self)
@@ -78,7 +95,7 @@ class S3Artifact(V1alpha1S3Artifact):
         return art
 
     def to_dict(self):
-        d = {"key": self.key, "urn": self.urn}
+        d = {"key": self.key, "urn": self.urn, "slice": self.slice}
         if s3_config["storage_client"] is None:
             d.update(s3_config)
         else:
@@ -87,7 +104,15 @@ class S3Artifact(V1alpha1S3Artifact):
 
     @classmethod
     def from_dict(cls, d):
-        return cls(key=d["key"], urn=d.get("urn", ""))
+        artifact = cls(key=d["key"], urn=d.get("urn", ""))
+        artifact.slice = d.get("slice")
+        return artifact
+
+    def __getstate__(self):
+        return self.to_dict()
+
+    def __setstate__(self, state):
+        self.__dict__.update(self.from_dict(state).__dict__)
 
     def sub_path(
             self,
@@ -96,7 +121,7 @@ class S3Artifact(V1alpha1S3Artifact):
         artifact = deepcopy(self)
         if artifact.key[-1:] != "/":
             artifact.key += "/"
-        artifact.key += path
+        artifact.key += str(path)
         return artifact
 
     def download(self, **kwargs):
@@ -109,6 +134,13 @@ class S3Artifact(V1alpha1S3Artifact):
         return V1alpha1OSSArtifact(key=s3_config["repo_prefix"] + self.key,
                                    local_vars_configuration=config)
 
+    def evalable_repr(self, imports):
+        args = "key='%s'" % self.key
+        if self.urn:
+            args += ", urn='%s'" % self.urn
+        imports.add(("dflow", "S3Artifact"))
+        return "S3Artifact(%s)" % args
+
 
 class LocalArtifact:
     def __init__(self, local_path):
@@ -125,6 +157,14 @@ class LocalArtifact:
             art.slice = "%s.%s" % (art.slice, key)
         return art
 
+    def sub_path(
+            self,
+            path: str,
+    ) -> Any:
+        artifact = deepcopy(self)
+        artifact.local_path += "/%s" % path
+        return artifact
+
 
 class HTTPArtifact:
     def __init__(self, url):
