@@ -35,7 +35,7 @@ config = {
     "k8s_config_file": os.environ.get("DFLOW_K8S_CONFIG_FILE", None),
     "k8s_api_server": os.environ.get("DFLOW_K8S_API_SERVER", None),
     "private_key_host_path": os.environ.get("DFLOW_PRIVATE_KEY_HOST_PATH",
-                                            "/home/docker/.ssh"),
+                                            None),
     "save_path_as_parameter": boolize(os.environ.get(
         "DFLOW_SAVE_PATH_AS_PARAMETER", False)),
     "catalog_dir_name": os.environ.get("DFLOW_CATALOG_DIR_NAME", ".dflow"),
@@ -65,9 +65,16 @@ config = {
     "debug_copy_method": os.environ.get("DFLOW_DEBUG_COPY_METHOD", "symlink"),
     "debug_pool_workers": (lambda s: None if s is None else int(s))(
         os.environ.get("DFLOW_DEBUG_POOL_WORKERS", None)),
+    "debug_batch_size": (lambda s: None if s is None else int(s))(
+        os.environ.get("DFLOW_DEBUG_BATCH_SIZE", None)),
+    "debug_batch_interval": int(os.environ.get("DFLOW_DEBUG_BATCH_INTERVAL",
+                                               30)),
     "detect_empty_dir": boolize(os.environ.get("DFLOW_DETECT_EMPTY_DIR",
                                                True)),
     "artifact_register": {},
+    "debug_s3": boolize(os.environ.get("DFLOW_DEBUG_S3", False)),
+    "debug_workdir": os.environ.get("DFLOW_DEBUG_WORKDIR", "."),
+    "debug_artifact_dir": os.environ.get("DFLOW_DEBUG_ARTIFACT_DIR", "."),
 }
 
 
@@ -113,7 +120,7 @@ s3_config = {
     "secure": boolize(os.environ.get("DFLOW_S3_SECURE", False)),
     "bucket_name": os.environ.get("DFLOW_S3_BUCKET_NAME", "my-bucket"),
     "repo_key": os.environ.get("DFLOW_S3_REPO_KEY", None),
-    "repo": os.environ.get("DFLOW_S3_REPO", None),
+    "repo": json.loads(os.environ.get("DFLOW_S3_REPO", "null")),
     "repo_type": os.environ.get("DFLOW_S3_REPO_TYPE", "s3"),
     "repo_prefix": os.environ.get("DFLOW_S3_REPO_PREFIX", ""),
     "prefix": os.environ.get("DFLOW_S3_PREFIX", ""),
