@@ -7,7 +7,8 @@ from typing import Any, Dict, List, Optional, Union
 
 import jsonpickle
 
-from .common import CustomArtifact, S3Artifact, param_errmsg, param_regex
+from .common import (CustomArtifact, LocalArtifact, S3Artifact, param_errmsg,
+                     param_regex)
 from .config import config
 from .utils import randstr, s3_config, upload_s3
 
@@ -22,6 +23,16 @@ except Exception:
 NotAllowedInputArtifactPath = ["/", "/tmp"]
 
 
+def type_to_str(type):
+    if hasattr(type, "__module__") and hasattr(type, "__name__"):
+        if type.__module__ == "builtins":
+            return type.__name__
+        else:
+            return "%s.%s" % (type.__module__, type.__name__)
+    else:
+        return str(type)
+
+
 class ObjectDict(UserDict):
     def __getattr__(self, key):
         if key == "data":
@@ -104,6 +115,10 @@ class AutonamedDict(UserDict):
         for value in self.values():
             value.template = template
 
+    def convert_to_graph(self):
+        return {k: v.convert_to_graph() for k, v in self.items()
+                if not k.startswith("dflow_")}
+
 
 class InputParameters(AutonamedDict):
     def __setitem__(self, key, value):
@@ -189,8 +204,11 @@ class OutputArtifacts(AutonamedDict):
 
 
 class ArgoVar:
-    def __init__(self, expr=None):
+    is_str = True
+
+    def __init__(self, expr=None, is_str=True):
         self.expr = expr
+        self.is_str = is_str
 
     def __repr__(self):
         return self.expr
@@ -205,8 +223,7 @@ class ArgoVar:
             item = "jsonpath(%s, '$')['%s']" % (self.expr, i)
         else:
             item = "jsonpath(%s, '$')[%s]" % (self.expr, i)
-        return ArgoVar("string(%s) == %s ? %s : toJson(%s)" % (
-            item, item, item, item))
+        return ArgoVar(item, is_str=False)
 
     def __iter__(self):
         raise TypeError("'ArgoVar' object is not iterable")
@@ -375,8 +392,10 @@ def convert_value_to_str(value):
                     if (hasattr(v, "type") and v.type == str) or (
                             hasattr(v, "value") and isinstance(v.value, str)):
                         vars.append("\"{{=%s}}\"" % v.expr)
-                    else:
+                    elif v.is_str:
                         vars.append("{{=%s}}" % v.expr)
+                    else:
+                        vars.append("{{=toJson(%s)}}" % v.expr)
                 elif isinstance(v, (dict, list)):
                     handle(v)
 
@@ -429,8 +448,9 @@ class InputParameter(ArgoVar):
     def from_dict(cls, d):
         kwargs = {
             "name": d.get("name", None),
+            # for backward compatible
             "type": str if json.loads(d.get("description", "{}")).get(
-                "type", str(str)) == str(str) else None,
+                "type", "str") in ["str", str(str)] else None,
         }
         if "value" in d:
             kwargs["value"] = d["value"]
@@ -491,11 +511,10 @@ class InputParameter(ArgoVar):
     def convert_to_argo(self):
         description = None
         if self.type is not None:
-            description = jsonpickle.dumps({"type": str(self.type)})
+            description = jsonpickle.dumps({"type": type_to_str(self.type)})
 
         if self.save_as_artifact:
-            if isinstance(self.source, (InputParameter, OutputParameter,
-                                        InputArtifact, OutputArtifact)):
+            if self.source is not None:
                 return V1alpha1Artifact(name="dflow_bigpar_" + self.name,
                                         path=self.path, _from=str(self.source))
             elif hasattr(self, "value"):
@@ -526,14 +545,50 @@ class InputParameter(ArgoVar):
         if not hasattr(self, "value"):
             return V1alpha1Parameter(name=self.name, description=description)
         elif isinstance(self.value, ArgoVar):
+            if self.value.is_str:
+                value = "{{=%s}}" % self.value.expr
+            else:
+                value = "{{=string(%s) == %s ? %s : toJson(%s)}}" % (
+                    self.value.expr, self.value.expr, self.value.expr,
+                    self.value.expr)
             return V1alpha1Parameter(name=self.name,
-                                     value="{{=%s}}" % self.value.expr,
+                                     value=value,
                                      description=description)
         else:
             return V1alpha1Parameter(name=self.name,
                                      value=convert_value_to_str(self.value),
                                      description=description)
 
+    def convert_to_graph(self):
+        g = {
+            "name": self.name,
+            "type": type_to_str(self.type) if self.type is not None else None,
+            "save_as_artifact": self.save_as_artifact,
+            "path": self.path,
+        }
+        if hasattr(self, "value"):
+            if isinstance(self.value, (InputParameter, OutputParameter)):
+                g["value"] = str(self.value)
+            elif isinstance(self.value, ArgoVar):
+                if self.value.is_str:
+                    g["value"] = "{{=%s}}" % self.value.expr
+                else:
+                    g["value"] = "{{=string(%s) == %s ? %s : toJson(%s)}}" % (
+                        self.value.expr, self.value.expr, self.value.expr,
+                        self.value.expr)
+            else:
+                g["value"] = self.value
+        elif self.source is not None:
+            if hasattr(self.source, "save_as_artifact"):
+                self.source.save_as_artifact = False
+            g["value"] = str(self.source) if isinstance(
+                self.source, ArgoVar) else self.source
+        return g
+
+    @classmethod
+    def from_graph(cls, graph):
+        return cls(**graph)
+
 
 class InputArtifact(ArgoVar):
     """
@@ -592,9 +647,9 @@ class InputArtifact(ArgoVar):
             "archive": None if "none" in d.get("archive", {}) else "tar",
         }
         if "s3" in d:
-            kwargs["source"] = S3Artifact(key=d["s3"]["key"])
+            kwargs["source"] = S3Artifact(key=d["s3"]["key"], debug_s3=True)
         if "oss" in d:
-            kwargs["source"] = S3Artifact(key=d["oss"]["key"])
+            kwargs["source"] = S3Artifact(key=d["oss"]["key"], debug_s3=True)
         return cls(**kwargs)
 
     def __getitem__(self, key):
@@ -652,9 +707,9 @@ class InputArtifact(ArgoVar):
         artifact = deepcopy(self)
         artifact.parent = self
         if artifact._sub_path is None:
-            artifact._sub_path = path
+            artifact._sub_path = str(path)
         else:
-            artifact._sub_path += "/" + path
+            artifact._sub_path += "/%s" % path
         return artifact
 
     def get_path_list_parameter(self):
@@ -688,9 +743,10 @@ class InputArtifact(ArgoVar):
             return V1alpha1Artifact(name=self.name, path=self.path,
                                     optional=self.optional, mode=self.mode,
                                     archive=archive)
-        if isinstance(self.source, (InputArtifact, OutputArtifact)):
+        if isinstance(self.source, (InputArtifact, OutputArtifact)) or (
+                isinstance(self.source, str) and self.source.find("{{") != -1):
             sub_path = self.sp if self.sp is not None else \
-                self.source._sub_path
+                getattr(self.source, "_sub_path", None)
             if isinstance(sub_path, ArgoVar):
                 sub_path = "{{=%s}}" % sub_path.expr
             return V1alpha1Artifact(name=self.name, path=self.path,
@@ -714,11 +770,40 @@ class InputArtifact(ArgoVar):
                                     optional=self.optional,
                                     raw=V1alpha1RawArtifact(data=self.source),
                                     mode=self.mode, archive=archive)
+        elif isinstance(self.source, LocalArtifact):
+            return {"name": self.name, "path": self.path,
+                    "optional": self.optional, "local": {
+                        "path": self.source.local_path}}
         else:
             raise RuntimeError(
                 "Cannot pass an object of type %s to artifact %s" %
                 (type(self.source), self))
 
+    def convert_to_graph(self):
+        if hasattr(self.source, "save_as_parameter"):
+            self.source.save_as_parameter = False
+        if hasattr(self.source, "redirect"):
+            self.source.redirect = None
+
+        source = str(self.source) if isinstance(
+            self.source, ArgoVar) else self.source
+        return {
+            "name": self.name,
+            "type": type_to_str(self.type) if self.type is not None else None,
+            "path": self.path,
+            "optional": self.optional,
+            "source": source,
+            "mode": self.mode,
+            "sub_path": self.sp,
+            "slice": self.slice,
+            "archive": self.archive,
+            "save_as_parameter": self.save_as_parameter,
+        }
+
+    @classmethod
+    def from_graph(cls, graph):
+        return cls(**graph)
+
 
 class OutputParameter(ArgoVar):
     """
@@ -772,12 +857,13 @@ class OutputParameter(ArgoVar):
         kwargs = {
             "name": d.get("name", None),
             "value_from_path": d.get("valueFrom", {}).get("path", None),
+            # for backward compatible
             "type": str if json.loads(d.get("description", "{}")).get(
-                "type", str(str)) == str(str) else None,
+                "type", "str") in ["str", str(str)] else None,
             "global_name": d.get("globalName", None),
             "value_from_expression": d.get("valueFrom", {}).get("expression",
                                                                 None),
-            "value_from_parameter": d.get("valueFrom", {}).get("parameters",
+            "value_from_parameter": d.get("valueFrom", {}).get("parameter",
                                                                None),
         }
         if "valueFrom" in d and "default" in d["valueFrom"]:
@@ -908,7 +994,7 @@ class OutputParameter(ArgoVar):
     def convert_to_argo_parameter(self):
         description = None
         if self.type is not None:
-            description = jsonpickle.dumps({"type": str(self.type)})
+            description = jsonpickle.dumps({"type": type_to_str(self.type)})
 
         default = None
         if hasattr(self, "default"):
@@ -960,6 +1046,32 @@ class OutputParameter(ArgoVar):
         else:
             return self.convert_to_argo_parameter()
 
+    def convert_to_graph(self):
+        g = {
+            "name": self.name,
+            "value_from_path": self.value_from_path,
+            "type": type_to_str(self.type) if self.type is not None else None,
+            "global_name": self.global_name,
+            "value_from_expression": str(self.value_from_expression)
+            if self.value_from_expression is not None else None,
+            "save_as_artifact": self.save_as_artifact,
+            "save_both": self.save_both,
+            "value_from_parameter": None,
+        }
+        if self.value_from_parameter is not None:
+            if hasattr(self.value_from_parameter, "save_as_artifact"):
+                self.value_from_parameter.save_as_artifact = False
+            g["value_from_parameter"] = str(self.value_from_parameter)
+        if hasattr(self, "default"):
+            g["default"] = self.default
+        if hasattr(self, "value"):
+            g["value"] = self.value
+        return g
+
+    @classmethod
+    def from_graph(cls, graph):
+        return cls(**graph)
+
 
 class OutputArtifact(ArgoVar):
     """
@@ -989,6 +1101,7 @@ class OutputArtifact(ArgoVar):
             archive: str = "default",
             global_name: Optional[str] = None,
             from_expression: Union[IfExpression, str] = None,
+            optional: bool = False,
     ) -> None:
         self.path = path
         self.name = name
@@ -1010,6 +1123,7 @@ class OutputArtifact(ArgoVar):
         self._from = _from
         self.slice = None
         self.parent = None
+        self.optional = optional
 
     @classmethod
     def from_dict(cls, d):
@@ -1020,6 +1134,7 @@ class OutputArtifact(ArgoVar):
             "global_name": d.get("globalName", None),
             "from_expression": d.get("fromExpression", None),
             "_from": d.get("from", None),
+            "optional": d.get("optional", False),
         }
         if "s3" in d:
             kwargs["save"] = [S3Artifact(key=d["s3"]["key"])]
@@ -1040,9 +1155,9 @@ class OutputArtifact(ArgoVar):
         artifact = deepcopy(self)
         artifact.parent = self
         if artifact._sub_path is None:
-            artifact._sub_path = path
+            artifact._sub_path = str(path)
         else:
-            artifact._sub_path += "/" + path
+            artifact._sub_path += "/%s" % path
         return artifact
 
     def __getattr__(self, key):
@@ -1143,7 +1258,8 @@ class OutputArtifact(ArgoVar):
     def convert_to_argo(self):
         kwargs = {
             "name": self.name,
-            "global_name": self.global_name
+            "global_name": self.global_name,
+            "optional": self.optional,
         }
 
         if self.archive is None:
@@ -1174,6 +1290,26 @@ class OutputArtifact(ArgoVar):
         else:
             raise RuntimeError("Output artifact %s is not specified" % self)
 
+    def convert_to_graph(self):
+        if hasattr(self._from, "redirect"):
+            self._from.redirect = None
+        return {
+            "name": self.name,
+            "path": self.path,
+            "type": type_to_str(self.type) if self.type is not None else None,
+            "save": self.save,
+            "archive": self.archive,
+            "global_name": self.global_name,
+            "from_expression": str(self.from_expression)
+            if self.from_expression is not None else None,
+            "_from": str(self._from) if self._from is not None else None,
+            "optional": self.optional,
+        }
+
+    @classmethod
+    def from_graph(cls, graph):
+        return cls(**graph)
+
 
 class Inputs:
     """
@@ -1252,6 +1388,22 @@ class Inputs:
                 artifacts.append(art.convert_to_argo())
         return V1alpha1Inputs(parameters=parameters, artifacts=artifacts)
 
+    def convert_to_graph(self):
+        return {
+            "parameters": self.parameters.convert_to_graph(),
+            "artifacts": self.artifacts.convert_to_graph(),
+        }
+
+    @classmethod
+    def from_graph(cls, graph):
+        graph["parameters"] = {
+            name: InputParameter.from_graph(par)
+            for name, par in graph.get("parameters", {}).items()}
+        graph["artifacts"] = {
+            name: InputArtifact.from_graph(art)
+            for name, art in graph.get("artifacts", {}).items()}
+        return cls(**graph)
+
 
 class Outputs:
     """
@@ -1330,3 +1482,19 @@ class Outputs:
         for art in self.artifacts.values():
             artifacts.append(art.convert_to_argo())
         return V1alpha1Outputs(parameters=parameters, artifacts=artifacts)
+
+    def convert_to_graph(self):
+        return {
+            "parameters": self.parameters.convert_to_graph(),
+            "artifacts": self.artifacts.convert_to_graph(),
+        }
+
+    @classmethod
+    def from_graph(cls, graph):
+        graph["parameters"] = {
+            name: OutputParameter.from_graph(par)
+            for name, par in graph.get("parameters", {}).items()}
+        graph["artifacts"] = {
+            name: OutputArtifact.from_graph(art)
+            for name, art in graph.get("artifacts", {}).items()}
+        return cls(**graph)
