@@ -1,9 +1,10 @@
 import argparse
 import datetime
+import json
 from typing import List, Optional
 
-from dflow import (S3Artifact, Workflow, config, download_artifact,
-                   query_workflows, upload_artifact)
+from dflow import (S3Artifact, Secret, Workflow, config, download_artifact,
+                   gen_code, query_workflows, upload_artifact)
 
 
 def main_parser():
@@ -111,6 +112,13 @@ def main_parser():
         formatter_class=argparse.ArgumentDefaultsHelpFormatter,
     )
     parser_retry.add_argument("ID", help="the workflow ID.")
+    parser_retry.add_argument(
+        "-s",
+        "--step",
+        type=str,
+        default=None,
+        help="retry a step in a running workflow with step ID (experimental)",
+    )
 
     parser_stop = subparsers.add_parser(
         "stop",
@@ -234,6 +242,49 @@ def main_parser():
         action="store_true",
         help="detach mode for running workflow",
     )
+    parser_submit.add_argument(
+        "-n",
+        "--name",
+        type=str,
+        default=None,
+        help="workflow ID",
+    )
+
+    parser_create = subparsers.add_parser(
+        "create",
+        help="Create a resource",
+        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
+    )
+    create_subparsers = parser_create.add_subparsers(
+        title="Valid resources", dest="resource")
+    parser_secret = create_subparsers.add_parser(
+        "secret",
+        help="Create a secret",
+        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
+    )
+    parser_secret.add_argument("value", help="secret value")
+    parser_secret.add_argument(
+        "-n",
+        "--name",
+        type=str,
+        help="name of the secret",
+    )
+    parser_secret.add_argument(
+        "-k",
+        "--key",
+        type=str,
+        default="secret",
+        help="key in the secret",
+    )
+
+    parser_codegen = subparsers.add_parser(
+        "codegen",
+        help="Generate code from graph",
+        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
+    )
+    parser_codegen.add_argument("GRAPH", help="the graph JSON file.")
+    parser_codegen.add_argument("-o", "--output", type=str,
+                                help="File path of the generated code")
     return parser
 
 
@@ -394,7 +445,11 @@ def main():
     elif args.command == "retry":
         wf_id = args.ID
         wf = Workflow(id=wf_id)
-        wf.retry()
+        if args.step is not None:
+            step = wf.query_step(id=args.step)[0]
+            step.retry()
+        else:
+            wf.retry()
     elif args.command == "stop":
         wf_id = args.ID
         wf = Workflow(id=wf_id)
@@ -443,9 +498,25 @@ def main():
     elif args.command == "submit":
         with open(args.FILE, "r") as f:
             wf = Workflow.from_yaml(f.read())
+        if args.name is not None:
+            wf.id = args.name
         if args.detach:
             config["detach"] = True
         wf.submit()
+    elif args.command == "create":
+        if args.resource == "secret":
+            s = Secret(args.value, args.name, args.key)
+            print("Secret (name: %s, key: %s) created" % (s.secret_name,
+                                                          s.secret_key))
+    elif args.command == "codegen":
+        with open(args.GRAPH, "r") as f:
+            graph = json.load(f)
+        code = gen_code(graph)
+        if args.output is None:
+            print(code)
+        else:
+            with open(args.output, "w") as f:
+                f.write(code)
 
 
 if __name__ == "__main__":
