@@ -9,46 +9,54 @@ from .io import PVC, InputParameter, Inputs, OutputParameter, Outputs
 from .utils import randstr
 
 try:
-    from argo.workflows.client import (V1alpha1Cache, V1alpha1Memoize,
-                                       V1alpha1Metadata,
+    from argo.workflows.client import (V1Affinity, V1alpha1Cache,
+                                       V1alpha1Memoize, V1alpha1Metadata,
                                        V1alpha1ResourceTemplate,
                                        V1alpha1ScriptTemplate,
                                        V1alpha1Template, V1alpha1UserContainer,
                                        V1ConfigMapKeySelector, V1EnvVar,
                                        V1EnvVarSource, V1ResourceRequirements,
-                                       V1SecretKeySelector, V1Volume,
-                                       V1VolumeMount)
+                                       V1SecretKeySelector, V1Toleration,
+                                       V1Volume, V1VolumeMount)
     from argo.workflows.client.configuration import Configuration
 
     import kubernetes
 
     from .client.v1alpha1_retry_strategy import V1alpha1RetryStrategy
 except Exception:
+    V1Affinity = object
     V1alpha1ResourceTemplate = object
     V1alpha1RetryStrategy = object
-    V1Volume = object
-    V1VolumeMount = object
     V1alpha1UserContainer = object
     V1EnvVarSource = object
+    V1Toleration = object
+    V1Volume = object
+    V1VolumeMount = object
 
 
 class Secret:
-    def __init__(self, value):
+    def __init__(self, value=None, name=None, key="secret"):
+        self.secret_name = "dflow-" + randstr(15) if name is None else name
+        self.secret_key = key
         self.value = value
-        self.secret_name = "dflow-" + randstr(15)
-        self.secret_key = "secret"
-        data = {self.secret_key: value}
-        secret = kubernetes.client.V1Secret(
-            string_data=data,
-            metadata=kubernetes.client.V1ObjectMeta(name=self.secret_name),
-            type="Opaque")
-        with get_k8s_client() as k8s_client:
-            core_v1_api = kubernetes.client.CoreV1Api(k8s_client)
-            core_v1_api.api_client.call_api(
-                '/api/v1/namespaces/%s/secrets' % global_config["namespace"],
-                'POST', body=secret, response_type='V1Secret',
-                header_params=global_config["http_headers"],
-                _return_http_data_only=True)
+        if value is not None:
+            data = {self.secret_key: value}
+            secret = kubernetes.client.V1Secret(
+                string_data=data,
+                metadata=kubernetes.client.V1ObjectMeta(name=self.secret_name),
+                type="Opaque")
+            with get_k8s_client() as k8s_client:
+                core_v1_api = kubernetes.client.CoreV1Api(k8s_client)
+                core_v1_api.api_client.call_api(
+                    '/api/v1/namespaces/%s/secrets' %
+                    global_config["namespace"],
+                    'POST', body=secret, response_type='V1Secret',
+                    header_params=global_config["http_headers"],
+                    _return_http_data_only=True)
+
+    def __repr__(self):
+        return "Secret(name='%s', key='%s')" % (self.secret_name,
+                                                self.secret_key)
 
 
 def get_k8s_client(k8s_api_server=None, token=None, k8s_config_file=None):
@@ -83,6 +91,7 @@ class OPTemplate:
             memoize_key: Optional[str] = None,
             pvcs: Optional[List[PVC]] = None,
             annotations: Dict[str, str] = None,
+            labels: Dict[str, str] = None,
     ) -> None:
         if name is None:
             name = randstr()
@@ -107,6 +116,7 @@ class OPTemplate:
         if annotations is None:
             annotations = {}
         self.annotations = annotations
+        self.labels = labels if labels is not None else {}
         self.modified = False
 
     @classmethod
@@ -206,6 +216,10 @@ class ScriptOPTemplate(OPTemplate):
         pvcs: PVCs need to be declared
         image_pull_policy: Always, IfNotPresent, Never
         annotations: annotations for the OP template
+        labels: labels for the OP template
+        node_selector: node selector when scheduling the pod
+        tolerations: tolerations of taints when scheduling the pod
+        affinity: affinity when scheduling the pod
         requests: a dict of resource requests
         limits: a dict of resource limits
         envs: environment variables
@@ -219,6 +233,10 @@ class ScriptOPTemplate(OPTemplate):
             memoize_key: Optional[str] = None,
             pvcs: Optional[List[PVC]] = None,
             annotations: Dict[str, str] = None,
+            labels: Dict[str, str] = None,
+            node_selector: Dict[str, str] = None,
+            tolerations: List[V1Toleration] = None,
+            affinity: V1Affinity = None,
             image: Optional[str] = None,
             command: Union[str, List[str]] = None,
             script: Optional[str] = None,
@@ -234,10 +252,11 @@ class ScriptOPTemplate(OPTemplate):
             envs: Dict[str, Union[str, Secret, V1EnvVarSource]] = None,
             init_containers: Optional[List[V1alpha1UserContainer]] = None,
             sidecars: Optional[List[V1alpha1UserContainer]] = None,
+            script_rendered: bool = False,
     ) -> None:
         super().__init__(name=name, inputs=inputs, outputs=outputs,
                          memoize_key=memoize_key, pvcs=pvcs,
-                         annotations=annotations)
+                         annotations=annotations, labels=labels)
         self.image = image
         if isinstance(command, str):
             command = [command]
@@ -263,7 +282,10 @@ class ScriptOPTemplate(OPTemplate):
         if sidecars is None:
             sidecars = []
         self.sidecars = sidecars
-        self.script_rendered = False
+        self.node_selector = node_selector if node_selector is not None else {}
+        self.tolerations = tolerations if tolerations is not None else []
+        self.affinity = affinity
+        self.script_rendered = script_rendered
 
     @classmethod
     def from_dict(cls, d):
@@ -289,16 +311,22 @@ class ScriptOPTemplate(OPTemplate):
                 "requests", None),
             "limits": d.get("script", {}).get("resources", {}).get(
                 "limits", None),
-            "envs": {env.name: env.value for env in d.get("script", {}).get(
-                "env", [])},
+            "envs": {env["name"]: env["value"] for env in d.get(
+                "script", {}).get("env", [])},
             "init_containers": d.get("initContainers", None),
             "sidecars": d.get("sidecars", None),
+            "node_selector": d.get("nodeSelector", None),
+            "tolerations": d.get("tolerations", None),
+            "affinity": d.get("affinity", None),
         }
         engine = d.get("metadata", {}).get("annotations", {}).get(
             "workflow.dp.tech/container_engine")
         docker = "docker" if engine == "docker" else None
         singularity = "singularity" if engine == "singularity" else None
         podman = "podman" if engine == "podman" else None
+        if kwargs["annotations"].get("workflow.dp.tech/script_rendered") == \
+                "true":
+            kwargs["script_rendered"] = True
         if d.get("metadata", {}).get("annotations", {}).get(
                 "workflow.dp.tech/executor") == "dispatcher":
             host = kwargs["annotations"].get("workflow.dp.tech/host")
@@ -332,6 +360,8 @@ class ScriptOPTemplate(OPTemplate):
                         memoize_configmap="dflow"):
         self.handle_key(memoize_prefix, memoize_configmap)
         self.annotations["workflows.argoproj.io/progress"] = self.init_progress
+        if self.script_rendered:
+            self.annotations["workflow.dp.tech/script_rendered"] = "true"
         if self.resource is not None:
             return V1alpha1Template(name=self.name,
                                     metadata=V1alpha1Metadata(
@@ -370,7 +400,11 @@ class ScriptOPTemplate(OPTemplate):
             return \
                 V1alpha1Template(name=self.name,
                                  metadata=V1alpha1Metadata(
-                                     annotations=self.annotations),
+                                     annotations=self.annotations,
+                                     labels=self.labels),
+                                 node_selector=self.node_selector,
+                                 tolerations=self.tolerations,
+                                 affinity=self.affinity,
                                  inputs=self.inputs.convert_to_argo(),
                                  outputs=self.outputs.convert_to_argo(),
                                  timeout=self.timeout,
@@ -391,6 +425,42 @@ class ScriptOPTemplate(OPTemplate):
                                  archive_location=loc,
                                  )
 
+    def convert_to_graph(self):
+        return {
+            "name": self.name,
+            "type": "ScriptOPTemplate",
+            "inputs": self.inputs.convert_to_graph(),
+            "outputs": self.outputs.convert_to_graph(),
+            "pvcs": self.pvcs,
+            "annotations": self.annotations,
+            "labels": self.labels,
+            "image": self.image,
+            "command": self.command,
+            "script": self.script,
+            "volumes": self.volumes,
+            "mounts": self.mounts,
+            "init_progress": self.init_progress,
+            "timeout": self.timeout,
+            "retry_strategy": self.retry_strategy,
+            "resource": self.resource,
+            "image_pull_policy": self.image_pull_policy,
+            "requests": self.requests,
+            "limits": self.limits,
+            "envs": self.envs,
+            "init_containers": self.init_containers,
+            "sidecars": self.sidecars,
+            "node_selector": self.node_selector,
+            "tolerations": self.tolerations,
+            "affinity": self.affinity,
+        }
+
+    @classmethod
+    def from_graph(cls, graph):
+        assert graph.pop("type") == "ScriptOPTemplate"
+        graph["inputs"] = Inputs.from_graph(graph.get("inputs", {}))
+        graph["outputs"] = Outputs.from_graph(graph.get("outputs", {}))
+        return cls(**graph)
+
 
 class ShellOPTemplate(ScriptOPTemplate):
     """
@@ -412,6 +482,10 @@ class ShellOPTemplate(ScriptOPTemplate):
         pvcs: PVCs need to be declared
         image_pull_policy: Always, IfNotPresent, Never
         annotations: annotations for the OP template
+        labels: labels for the OP template
+        node_selector: node selector when scheduling the pod
+        tolerations: tolerations of taints when scheduling the pod
+        affinity: affinity when scheduling the pod
         requests: a dict of resource requests
         limits: a dict of resource limits
         envs: environment variables
@@ -427,6 +501,10 @@ class ShellOPTemplate(ScriptOPTemplate):
         memoize_key: Optional[str] = None,
         pvcs: Optional[List[PVC]] = None,
         annotations: Dict[str, str] = None,
+        labels: Dict[str, str] = None,
+        node_selector: Dict[str, str] = None,
+        tolerations: List[V1Toleration] = None,
+        affinity: V1Affinity = None,
         image: Optional[str] = None,
         command: Union[str, List[str]] = None,
         script: Optional[str] = None,
@@ -450,8 +528,9 @@ class ShellOPTemplate(ScriptOPTemplate):
             init_progress=init_progress, timeout=timeout,
             retry_strategy=retry_strategy, memoize_key=memoize_key, pvcs=pvcs,
             image_pull_policy=image_pull_policy, annotations=annotations,
-            requests=requests, limits=limits, envs=envs,
-            init_containers=init_containers, sidecars=sidecars,
+            labels=labels, node_selector=node_selector, affinity=affinity,
+            tolerations=tolerations, requests=requests, limits=limits,
+            envs=envs, init_containers=init_containers, sidecars=sidecars,
         )
 
 
@@ -475,6 +554,10 @@ class PythonScriptOPTemplate(ScriptOPTemplate):
         pvcs: PVCs need to be declared
         image_pull_policy: Always, IfNotPresent, Never
         annotations: annotations for the OP template
+        labels: labels for the OP template
+        node_selector: node selector when scheduling the pod
+        tolerations: tolerations of taints when scheduling the pod
+        affinity: affinity when scheduling the pod
         requests: a dict of resource requests
         limits: a dict of resource limits
         envs: environment variables
@@ -490,6 +573,10 @@ class PythonScriptOPTemplate(ScriptOPTemplate):
         memoize_key: Optional[str] = None,
         pvcs: Optional[List[PVC]] = None,
         annotations: Dict[str, str] = None,
+        labels: Dict[str, str] = None,
+        node_selector: Dict[str, str] = None,
+        tolerations: List[V1Toleration] = None,
+        affinity: V1Affinity = None,
         image: Optional[str] = None,
         command: Union[str, List[str]] = None,
         script: Optional[str] = None,
@@ -513,8 +600,9 @@ class PythonScriptOPTemplate(ScriptOPTemplate):
             init_progress=init_progress, timeout=timeout,
             retry_strategy=retry_strategy, memoize_key=memoize_key, pvcs=pvcs,
             image_pull_policy=image_pull_policy, annotations=annotations,
-            requests=requests, limits=limits, envs=envs,
-            init_containers=init_containers, sidecars=sidecars,
+            labels=labels, node_selector=node_selector, affinity=affinity,
+            tolerations=tolerations, requests=requests, limits=limits,
+            envs=envs, init_containers=init_containers, sidecars=sidecars,
         )
 
 
@@ -557,8 +645,8 @@ class ContainerOPTemplate(ScriptOPTemplate):
                 "requests", None),
             "limits": d.get("container", {}).get("resources", {}).get(
                 "limits", None),
-            "envs": {env.name: env.value for env in d.get("container", {}).get(
-                "env", [])},
+            "envs": {env["name"]: env["value"] for env in d.get(
+                "container", {}).get("env", [])},
             "init_containers": d.get("initContainers", None),
             "sidecars": d.get("sidecars", None),
         }
