@@ -19,6 +19,8 @@ except Exception:
 
 global_config["artifact_register"]["launching+datasets"] = \
     "dflow.plugins.datasets.DatasetsArtifact"
+global_config["artifact_register"]["launching+models"] = \
+    "dflow.plugins.datasets.DatasetsArtifact"
 
 
 def wait_for_mount(point, timeout=60):
@@ -130,9 +132,9 @@ class DatasetsArtifact(DispatcherArtifact):
     def sub_path(self, path: str):
         artifact = deepcopy(self)
         if artifact._sub_path is None:
-            artifact._sub_path = path
+            artifact._sub_path = str(path)
         else:
-            artifact._sub_path += "/" + path
+            artifact._sub_path += "/%s" % path
         return artifact
 
     def get_urn(self) -> str:
@@ -148,6 +150,11 @@ class DatasetsArtifact(DispatcherArtifact):
         return "launching://%s/%s@%s?alias=%s&action=%s" % (
             self.type, self.element, self.version, name, mount_type)
 
+    def modify_config(self, name: str, machine) -> str:
+        if "job_resources" not in machine.input_data:
+            machine.input_data["job_resources"] = []
+        machine.input_data["job_resources"].append(self.get_bohrium_urn(name))
+
     def download(self, name: str, path: str):
         wait_for_mount("/launching/%s" % name)
         if self._sub_path is not None:
@@ -166,6 +173,7 @@ class DatasetsArtifact(DispatcherArtifact):
         return p.pid
 
     def bohrium_download(self, name: str, path: str):
+        os.stat("/launching/%s" % name)
         if self._sub_path is not None:
             os.symlink("/launching/%s/%s" % (name, self._sub_path), path)
         else:
