@@ -14,7 +14,6 @@ from ..op_template import ScriptOPTemplate
 from ..python import PythonOPTemplate
 from ..python.python_op_template import handle_packages_script
 from ..utils import randstr, upload_s3
-from . import bohrium
 
 try:
     from argo.workflows.client import (V1HostPathVolumeSource, V1Volume,
@@ -77,6 +76,7 @@ class DispatcherExecutor(Executor):
                  json_file: Optional[os.PathLike] = None,
                  docker_executable: Optional[str] = None,
                  singularity_executable: Optional[str] = None,
+                 container_args: str = "",
                  podman_executable: Optional[str] = None,
                  remote_root: Optional[str] = None,
                  retry_on_submission_error: Optional[int] = None,
@@ -84,6 +84,7 @@ class DispatcherExecutor(Executor):
                  pre_script: str = "",
                  post_script: str = "",
                  clean: bool = True,
+                 remove_scheduling_strategies: bool = True,
                  ) -> None:
         self.host = host
         self.queue_name = queue_name
@@ -111,16 +112,15 @@ class DispatcherExecutor(Executor):
                 self.singularity_executable is not None or \
                 self.podman_executable is not None:
             self.map_tmp_dir = False
-        if config["mode"] == "debug":
-            self.work_root = "."
-        else:
-            self.work_root = "/"
+        self.container_args = container_args
+        self.work_root = "."
         self.remote_root = remote_root
         self.retry_on_submission_error = retry_on_submission_error
         self.merge_sliced_step = merge_sliced_step
         self.pre_script = pre_script
         self.post_script = post_script
         self.clean = clean
+        self.remove_scheduling_strategies = remove_scheduling_strategies
 
         conf = {}
         if json_file is not None:
@@ -157,6 +157,7 @@ class DispatcherExecutor(Executor):
             update_dict(self.machine_dict, machine_dict)
 
         if self.machine_dict["context_type"] == "Bohrium":
+            from . import bohrium
             if "batch_type" not in self.machine_dict:
                 self.machine_dict["batch_type"] = "Bohrium"
             self.bohrium_ticket = bohrium.config["ticket"]
@@ -230,11 +231,26 @@ class DispatcherExecutor(Executor):
         if not isinstance(template, ScriptOPTemplate):
             return template
 
+        machine_dict = deepcopy(self.machine_dict)
+        task_dict = deepcopy(self.task_dict)
         new_template = deepcopy(template)
         new_template.name += "-" + randstr()
         new_template.image = self.image
         new_template.image_pull_policy = self.image_pull_policy
         new_template.command = self.command
+        if self.remove_scheduling_strategies:
+            new_template.node_selector = None
+            new_template.tolerations = []
+            new_template.affinity = None
+            # dispatcher pod use ephemeral storage currently
+            if new_template.requests is not None:
+                new_template.requests = {
+                    k: v for k, v in new_template.requests.items()
+                    if k == "ephemeral-storage"}
+            if new_template.limits is not None:
+                new_template.limits = {
+                    k: v for k, v in new_template.limits.items()
+                    if k == "ephemeral-storage"}
 
         remote_command = template.command if self.remote_command is None \
             else self.remote_command
@@ -244,42 +260,44 @@ class DispatcherExecutor(Executor):
 
         cmd += run_script(template.image, remote_command,
                           self.docker_executable, self.singularity_executable,
-                          self.podman_executable)
-        self.task_dict["command"] = cmd
-        self.task_dict["forward_files"] = ["script"]
+                          self.podman_executable, args=self.container_args)
+        task_dict["command"] = cmd
+        task_dict["forward_files"] = ["script"]
         for art in template.inputs.artifacts.values():
-            self.task_dict["forward_files"].append("./" + art.path)
+            task_dict["forward_files"].append("./" + art.path)
         for par in template.inputs.parameters.values():
             if par.save_as_artifact:
-                self.task_dict["forward_files"].append("./" + par.path)
+                task_dict["forward_files"].append("./" + par.path)
         merge = self.merge_sliced_step and hasattr(template, "slices") and\
             template.slices is not None
         if merge:
             sliced_output_parameters = template.slices.output_parameter.copy()
             if "dflow_success_tag" in template.outputs.parameters:
                 sliced_output_parameters.append("dflow_success_tag")
-        self.task_dict["backward_files"] = []
+        task_dict["backward_files"] = []
         for art in template.outputs.artifacts.values():
-            self.task_dict["backward_files"].append("./" + art.path)
+            task_dict["backward_files"].append("./" + art.path)
         for name, par in template.outputs.parameters.items():
             if par.save_as_artifact:
-                self.task_dict["backward_files"].append(
+                task_dict["backward_files"].append(
                     "./" + par.value_from_path)
             elif par.value_from_path is not None and not (
                     merge and sliced_output_parameters):
-                self.task_dict["backward_files"].append(
+                task_dict["backward_files"].append(
                     "./" + par.value_from_path)
 
         new_template.script = self.pre_script
         new_template.script += "import os\n"
-        new_template.script += "os.chdir('%s')\n" % self.work_root
+        # compatible for debug mode
+        new_template.script += "if not os.getcwd().endswith('workdir'):\n"
+        new_template.script += "    os.chdir('/')\n"
         if any([art.save_as_parameter for art in
                 template.inputs.artifacts.values()]) and \
                 isinstance(template, PythonOPTemplate):
             template = deepcopy(template)
-            if self.machine_dict["context_type"] == "Bohrium":
+            if machine_dict["context_type"] == "Bohrium":
                 template.download_method = "bohrium_download"
-            elif self.machine_dict["context_type"] == "SSHContext":
+            elif machine_dict["context_type"] == "SSHContext":
                 template.download_method = "remote_download"
             template.render_script()
             new_template.volumes = [v for v in new_template.volumes
@@ -290,7 +308,6 @@ class DispatcherExecutor(Executor):
                                      if not s.name.startswith("rclone-")]
             new_template.script += handle_packages_script(
                 "./tmp/inputs/artifacts/dflow_python_packages")
-            new_template.script += "import jsonpickle\n"
             new_template.script += "from dflow import config\n"
             new_template.script += "config['artifact_register'] = %s\n" % \
                 json.dumps(config["artifact_register"])
@@ -305,33 +322,29 @@ class DispatcherExecutor(Executor):
         new_template.script += script
         new_template.script += "\"\"\")\n"
 
-        if self.machine_dict["context_type"] == "Bohrium":
+        if machine_dict["context_type"] == "Bohrium":
             if self.bohrium_ticket is not None:
                 if new_template.envs is None:
                     new_template.envs = {}
                 new_template.envs["BOHR_TICKET"] = self.bohrium_ticket
-            if "image_name" not in self.machine_dict["remote_profile"][
+            if "image_name" not in machine_dict["remote_profile"][
                     "input_data"]:
-                self.machine_dict["remote_profile"]["input_data"][
+                machine_dict["remote_profile"]["input_data"][
                     "image_name"] = template.image
 
-        self.machine_dict["local_root"] = self.work_root
+        machine_dict["local_root"] = self.work_root
         new_template.script += "import json, shlex\n"
         new_template.script += "from dpdispatcher import Machine, Resources,"\
             " Task, Submission\n"
         new_template.script += "machine = Machine.load_from_dict(json.loads("\
-            "r'%s'))\n" % json.dumps(self.machine_dict)
+            "r'%s'))\n" % json.dumps(machine_dict)
         for name, art in template.inputs.artifacts.items():
-            if self.machine_dict["context_type"] == "Bohrium" and \
+            if machine_dict["context_type"] == "Bohrium" and \
                     art.save_as_parameter:
-                new_template.script += "import jsonpickle\n"
                 new_template.script += "from dflow import CustomArtifact\n"
-                new_template.script += "bohrium_urn = CustomArtifact.from_urn"\
-                    "('{{inputs.parameters.dflow_art_%s}}').get_bohrium_urn("\
-                    "'%s')\n" % (name, name)
-                new_template.script += "machine.input_data['job_resources']"\
-                    " = machine.input_data.get('job_resources', []) + "\
-                    "[bohrium_urn]\n"
+                new_template.script += "CustomArtifact.from_urn"\
+                    "('{{inputs.parameters.dflow_art_%s}}').modify_config("\
+                    "'%s', machine)\n" % (name, name)
 
         new_template.script += "resources = Resources.load_from_dict(json."\
             "loads(r'%s'))\n" % json.dumps(self.resources_dict)
@@ -342,7 +355,7 @@ class DispatcherExecutor(Executor):
         new_template.script += "resources.envs['ARGO_TEMPLATE'] = "\
             "shlex.quote(os.environ.get('ARGO_TEMPLATE'))\n"
         new_template.script += "task = Task.load_from_dict(json.loads(r'%s'))"\
-            "\n" % json.dumps(self.task_dict)
+            "\n" % json.dumps(task_dict)
         new_template.script += "task.forward_files = list(filter("\
             "os.path.exists, task.forward_files))\n"
         new_template.script += "for f in task.backward_files:\n"
@@ -431,7 +444,16 @@ class DispatcherExecutor(Executor):
         else:
             new_template.script += "submission = Submission(work_base='.', "\
                 "machine=machine, resources=resources, task_list=[task])\n"
+        new_template.script += "def sigterm_handler(signum, frame):\n"
+        new_template.script += "    print('Got SIGTERM, kill unfinished tasks"\
+            "!')\n"
+        new_template.script += "    submission.remove_unfinished_tasks()\n"
+        new_template.script += "import signal\n"
+        new_template.script += "signal.signal(signal.SIGTERM, sigterm_handler"\
+            ")\n"
         if self.retry_on_submission_error:
+            new_template.script += "from dpdispatcher.dpcloudserver.client "\
+                "import RequestInfoException\n"
             new_template.script += "for retry in range(%s):\n" % \
                 self.retry_on_submission_error
             new_template.script += "    try:\n"
@@ -439,7 +461,7 @@ class DispatcherExecutor(Executor):
             new_template.script += "        submission.run_submission(clean="\
                 "%s)\n" % self.clean
             new_template.script += "        break\n"
-            new_template.script += "    except Exception:\n"
+            new_template.script += "    except RequestInfoException:\n"
             new_template.script += "        import traceback\n"
             new_template.script += "        traceback.print_exc()\n"
             new_template.script += "        import time\n"
@@ -460,7 +482,7 @@ class DispatcherExecutor(Executor):
                 new_template.script += "    f.write(json.dumps(res))\n"
         new_template.script_rendered = True
 
-        # workaround for unavailable exit code of Bohrium job
+        # workaround for unavailable exit code of dispatcher
         # check output files explicitly
         for art in template.outputs.artifacts.values():
             new_template.script += "assert os.path.exists('./%s')\n" % art.path
@@ -477,7 +499,7 @@ class DispatcherExecutor(Executor):
             new_template.inputs.artifacts["dflow_private_key"] = InputArtifact(
                 path="/root/.ssh/" + os.path.basename(self.private_key_file),
                 source=private_key_artifact, mode=0o600)
-        else:
+        elif config["private_key_host_path"] is not None:
             new_template.volumes.append(V1Volume(
                 name="dflow-private-key", host_path=V1HostPathVolumeSource(
                     path=config["private_key_host_path"])))
@@ -488,7 +510,7 @@ class DispatcherExecutor(Executor):
 
 class DispatcherArtifact(CustomArtifact, ABC):
     @abc.abstractmethod
-    def get_bohrium_urn(self, name: str) -> str:
+    def modify_config(self, name: str, machine) -> str:
         pass
 
     @abc.abstractmethod
