@@ -1,7 +1,6 @@
 import inspect
 import logging
 import os
-import sys
 from pathlib import Path
 from typing import Any, Dict, List, Optional, Type, Union
 
@@ -14,19 +13,21 @@ from ..config import config
 from ..io import (PVC, InputArtifact, InputParameter, Inputs, OutputArtifact,
                   OutputParameter, Outputs)
 from ..op_template import PythonScriptOPTemplate
-from ..utils import randstr, s3_config
-from .op import OP, iwd
+from ..utils import evalable_repr, randstr, s3_config
+from .op import OP, get_source_code
 from .opio import Artifact, BigParameter, Parameter
 
 try:
-    from argo.workflows.client import (V1alpha1UserContainer, V1Volume,
-                                       V1VolumeMount)
+    from argo.workflows.client import (V1Affinity, V1alpha1UserContainer,
+                                       V1Toleration, V1Volume, V1VolumeMount)
 
     from ..client import V1alpha1RetryStrategy
 except Exception:
+    V1Affinity = object
+    V1alpha1UserContainer = object
+    V1Toleration = object
     V1Volume = object
     V1VolumeMount = object
-    V1alpha1UserContainer = object
 upload_packages = []
 
 
@@ -61,9 +62,6 @@ class Slices:
             pool_size: Optional[int] = None,
             register_first_only: bool = False,
     ) -> None:
-        if sub_path:
-            assert input_artifact is not None, \
-                "input artifact not provided in subpath mode"
         self.input_parameter = input_parameter if input_parameter is not \
             None else []
         self.input_artifact = input_artifact if input_artifact is not None \
@@ -85,16 +83,23 @@ class Slices:
         self.pool_size = pool_size
         self.register_first_only = register_first_only
 
-
-def get_source_code(o):
-    source_lines, start_line = inspect.getsourcelines(o)
-    if sys.version_info.minor >= 9:
-        source_file = inspect.getsourcefile(o)
-    else:
-        source_file = os.path.join(iwd, inspect.getsourcefile(o))
-    with open(source_file, "r", encoding="utf-8") as fd:
-        pre_lines = fd.readlines()[:start_line-1]
-    return "".join(pre_lines + source_lines) + "\n"
+    def evalable_repr(self, imports):
+        kwargs = {}
+        sign = inspect.signature(self.__init__).parameters
+        for k, v in self.__dict__.items():
+            if k in sign:
+                if v == sign[k].default:
+                    continue
+                if sign[k].default is None and v in [[], {}]:
+                    continue
+                if k == "slices" and self.sub_path and v == "{{item.order}}":
+                    continue
+                if k == "slices" and not self.sub_path and v == "{{item}}":
+                    continue
+                kwargs[k] = v
+        imports.add(("dflow.python", "Slices"))
+        return "Slices(%s)" % ", ".join(["%s=%s" % (k, evalable_repr(
+            v, imports)) for k, v in kwargs.items()])
 
 
 def handle_packages_script(package_root):
@@ -123,6 +128,11 @@ class PythonOPTemplate(PythonScriptOPTemplate):
         op_class: Python class OP
         image: image of the OP template
         command: python executable
+        annotations: annotations for the OP template
+        labels: labels for the OP template
+        node_selector: node selector when scheduling the pod
+        tolerations: tolerations of taints when scheduling the pod
+        affinity: affinity when scheduling the pod
         input_artifact_slices: a dict specifying input artifacts to use slices
         output_artifact_save: a dict specifying storage of output artifacts
             overriding default storage
@@ -161,6 +171,11 @@ class PythonOPTemplate(PythonScriptOPTemplate):
                  op_class: Union[Type[OP], OP],
                  image: Optional[str] = None,
                  command: Union[str, List[str]] = None,
+                 annotations: Dict[str, str] = None,
+                 labels: Dict[str, str] = None,
+                 node_selector: Dict[str, str] = None,
+                 tolerations: List[V1Toleration] = None,
+                 affinity: V1Affinity = None,
                  output_artifact_save: Dict[str,
                                             List[Union[PVC, S3Artifact]]]
                  = None,
@@ -177,6 +192,9 @@ class PythonOPTemplate(PythonScriptOPTemplate):
                  python_packages: Optional[List[os.PathLike]] = None,
                  timeout: Optional[int] = None,
                  retry_on_transient_error: Optional[int] = None,
+                 retry_on_failure: Optional[int] = None,
+                 retry_on_error: Optional[int] = None,
+                 retry_on_failure_and_error: Optional[int] = None,
                  timeout_as_transient_error: bool = False,
                  memoize_key: Optional[str] = None,
                  volumes: Optional[List[V1Volume]] = None,
@@ -191,6 +209,7 @@ class PythonOPTemplate(PythonScriptOPTemplate):
                  tmp_root: str = "/tmp",
                  pre_script: str = "",
                  post_script: str = "",
+                 success_tag: bool = False,
                  ) -> None:
         self.n_parts = {}
         self.keys_of_parts = {}
@@ -215,12 +234,18 @@ class PythonOPTemplate(PythonScriptOPTemplate):
             name="%s-%s" % (class_name.replace("_", "-"), randstr()),
             inputs=Inputs(), outputs=Outputs(), volumes=volumes, mounts=mounts,
             requests=requests, limits=limits, envs=envs,
-            init_containers=init_containers, sidecars=sidecars)
+            init_containers=init_containers, sidecars=sidecars, labels=labels,
+            annotations=annotations, node_selector=node_selector,
+            tolerations=tolerations, affinity=affinity)
         self.pre_script = pre_script
         self.post_script = post_script
+        self.success_tag = success_tag
         if timeout is not None:
             self.timeout = "%ss" % timeout
         self.retry_on_transient_error = retry_on_transient_error
+        self.retry_on_failure = retry_on_failure
+        self.retry_on_error = retry_on_error
+        self.retry_on_failure_and_error = retry_on_failure_and_error
         self.timeout_as_transient_error = timeout_as_transient_error
         self.dflow_vars = {}
         self.tmp_root = tmp_root
@@ -255,7 +280,8 @@ class PythonOPTemplate(PythonScriptOPTemplate):
                 self.outputs.artifacts[name] = OutputArtifact(
                     path="%s/outputs/artifacts/" % self.tmp_root + name,
                     archive=sign.archive, save=sign.save,
-                    global_name=sign.global_name, type=sign.type)
+                    global_name=sign.global_name, type=sign.type,
+                    optional=sign.optional)
                 if config["save_path_as_parameter"]:
                     self.outputs.parameters["dflow_%s_path_list" %
                                             name] = OutputParameter(
@@ -308,6 +334,7 @@ class PythonOPTemplate(PythonScriptOPTemplate):
         else:
             python_packages = upload_packages + [python_packages]
 
+        self.upload_dflow = upload_dflow
         if upload_dflow:
             python_packages += __path__
             python_packages += jsonpickle.__path__
@@ -377,7 +404,6 @@ class PythonOPTemplate(PythonScriptOPTemplate):
             for name in slices.input_parameter:
                 self.input_parameter_slices[name] = slices.slices
         if slices.output_artifact:
-            self.output_artifact_slices = {}
             for name in slices.output_artifact:
                 self.output_artifact_slices[name] = slices.slices
                 self.outputs.artifacts[name].archive = None  # no archive
@@ -418,7 +444,11 @@ class PythonOPTemplate(PythonScriptOPTemplate):
             jsonpickle.dumps(config)
         script += "s3_config.update(jsonpickle.loads(r'''%s'''))\n" % \
             jsonpickle.dumps(s3_config)
-        if op_class.__module__ in ["__main__", "__mp_main__"]:
+        mod = op_class.__module__
+        if hasattr(op_class, "_source"):
+            script += op_class._source
+            mod = "__main__"
+        elif mod in ["__main__", "__mp_main__"]:
             try:
                 if hasattr(op_class, "func"):
                     script += get_source_code(op_class.func)
@@ -448,7 +478,7 @@ class PythonOPTemplate(PythonScriptOPTemplate):
                   " handle_input_parameter\n"
         script += "from dflow.python.utils import handle_output_artifact," \
                   " handle_output_parameter, handle_lineage\n"
-        script += f"from {op_class.__module__} import {class_name}\n\n"
+        script += f"from {mod} import {class_name}\n\n"
         if hasattr(op_class, "func"):
             script += "op_obj = %s\n" % class_name
         elif op is None:
@@ -494,8 +524,7 @@ class PythonOPTemplate(PythonScriptOPTemplate):
         for name, sign in input_sign.items():
             if isinstance(sign, Artifact):
                 slices = self.get_slices(input_artifact_slices, name)
-                if self.slices is not None and self.slices.sub_path and \
-                        name in self.slices.input_artifact:
+                if "dflow_%s_sub_path" % name in self.inputs.parameters:
                     script += "    input['%s'] = handle_input_artifact('%s', "\
                         "input_sign['%s'], %s, r'%s', '{{inputs.parameters."\
                         "dflow_%s_sub_path}}', None)\n" % (
@@ -524,18 +553,23 @@ class PythonOPTemplate(PythonScriptOPTemplate):
         if self.slices is not None and self.slices.pool_size is not None:
             sliced_inputs = self.slices.input_artifact + \
                 self.slices.input_parameter
-            if len(sliced_inputs) > 1:
-                script += "    assert %s\n" % " == ".join(
-                    ["len(input['%s'])" % i for i in sliced_inputs])
-            script += "    n_slices = len(input['%s'])\n" % \
-                sliced_inputs[0]
+            script += "    n_slices = None\n"
+            for name in sliced_inputs:
+                # for optional artifact
+                script += "    if input['%s'] is not None:\n" % name
+                script += "        if n_slices is None:\n"
+                script += "            n_slices = len(input['%s'])\n" % name
+                script += "        else:\n"
+                script += "            assert len(input['%s']) == n_slices\n" \
+                    % name
+            script += "    assert n_slices is not None\n"
             script += "    input_list = []\n"
             script += "    from copy import deepcopy\n"
             script += "    for i in range(n_slices):\n"
             script += "        input1 = deepcopy(input)\n"
             for name in sliced_inputs:
-                script += "        input1['%s'] = list(input['%s'])[i]\n" % (
-                    name, name)
+                script += "        input1['%s'] = list(input['%s'])[i] if "\
+                    "input['%s'] is not None else None\n" % (name, name, name)
             script += "        input_list.append(input1)\n"
             if self.slices.pool_size == 1:
                 script += "    output_list = []\n"
@@ -622,14 +656,25 @@ class PythonOPTemplate(PythonScriptOPTemplate):
                 "dflow_workflow_urn}}', r'%s')\n" % self.tmp_root
 
         if self.slices is not None and self.slices.pool_size is not None:
-            script += "    try:\n"
-            script += "        for error in error_list:\n"
-            script += "            if error is not None:\n"
-            script += "                raise error\n"
-            script += "    except TransientError:\n"
-            script += "        sys.exit(1)\n"
-            script += "    except FatalError:\n"
-            script += "        sys.exit(2)\n"
+            if self.success_tag:
+                script += "    with open('%s/outputs/success_tag', 'w') as f:"\
+                    "\n" % self.tmp_root
+                script += "        f.write(str(len([e for e in error_list if "\
+                    "e is None])))\n"
+            else:
+                script += "    try:\n"
+                script += "        for error in error_list:\n"
+                script += "            if error is not None:\n"
+                script += "                raise error\n"
+                script += "    except TransientError:\n"
+                script += "        sys.exit(1)\n"
+                script += "    except FatalError:\n"
+                script += "        sys.exit(2)\n"
+        else:
+            if self.success_tag:
+                script += "    with open('%s/outputs/success_tag', 'w') as f:"\
+                    "\n" % self.tmp_root
+                script += "        f.write('1')\n"
 
         script += self.post_script.format(**{"tmp_root": self.tmp_root})
 
@@ -673,8 +718,53 @@ class PythonOPTemplate(PythonScriptOPTemplate):
                 expr = "asInt(lastRetry.exitCode) == 1"
             self.retry_strategy = V1alpha1RetryStrategy(
                 limit=self.retry_on_transient_error, expression=expr)
+        elif self.retry_on_failure is not None:
+            self.retry_strategy = V1alpha1RetryStrategy(
+                limit=self.retry_on_failure, retry_policy="OnFailure")
+        elif self.retry_on_error is not None:
+            self.retry_strategy = V1alpha1RetryStrategy(
+                limit=self.retry_on_error, retry_policy="OnError")
+        elif self.retry_on_failure_and_error is not None:
+            self.retry_strategy = V1alpha1RetryStrategy(
+                limit=self.retry_on_failure_and_error, retry_policy="Always")
         return super().convert_to_argo(memoize_prefix, memoize_configmap)
 
+    def convert_to_graph(self):
+        g = super().convert_to_graph()
+        del g["name"]
+        del g["script"]
+        del g["inputs"]
+        del g["outputs"]
+        del g["pvcs"]
+        del g["init_progress"]
+        del g["retry_strategy"]
+        del g["resource"]
+        python_packages = []
+        for p in self.python_packages:
+            if self.upload_dflow and Path(p).name in [
+                    "dflow", "jsonpickle", "typeguard"]:
+                continue
+            python_packages.append(str(p))
+        g["type"] = "PythonOPTemplate"
+        g["op"] = self.op_class.convert_to_graph()
+        g["python_packages"] = python_packages
+        g["retry_on_transient_error"] = self.retry_on_transient_error
+        g["retry_on_failure"] = self.retry_on_failure
+        g["retry_on_error"] = self.retry_on_error
+        g["retry_on_failure_and_error"] = self.retry_on_failure_and_error
+        g["timeout_as_transient_error"] = self.timeout_as_transient_error
+        g["upload_dflow"] = self.upload_dflow
+        g["tmp_root"] = self.tmp_root
+        g["pre_script"] = self.pre_script
+        g["post_script"] = self.post_script
+        return g
+
+    @classmethod
+    def from_graph(cls, graph):
+        assert graph.pop("type") == "PythonOPTemplate"
+        graph["op_class"] = OP.from_graph(graph.pop("op"))
+        return cls(**graph)
+
 
 class TransientError(Exception):
     pass
