@@ -7,8 +7,9 @@ from typing import Dict, List, Set
 import jsonpickle
 
 from ..config import config
-from ..utils import (assemble_path_object, convert_dflow_list, copy_file,
-                     expand, flatten, randstr, remove_empty_dir_tag)
+from ..utils import (artifact_classes, assemble_path_object,
+                     convert_dflow_list, copy_file, expand, flatten, randstr,
+                     remove_empty_dir_tag)
 from .opio import Artifact, BigParameter, NestedDict, Parameter
 
 
@@ -31,6 +32,7 @@ def get_slices(path_object, slices):
 def handle_input_artifact(name, sign, slices=None, data_root="/tmp",
                           sub_path=None, n_parts=None, keys_of_parts=None,
                           path=None, prefix=None):
+    root = None
     if n_parts is not None:
         path_object = []
         for i in range(n_parts):
@@ -65,6 +67,7 @@ def handle_input_artifact(name, sign, slices=None, data_root="/tmp",
             if path is None else path
         if sub_path is not None:
             art_path = os.path.join(art_path, sub_path)
+        root = art_path
         if not os.path.exists(art_path):  # for optional artifact
             return None
         if config["detect_empty_dir"]:
@@ -83,10 +86,10 @@ def handle_input_artifact(name, sign, slices=None, data_root="/tmp",
             res = path_object[0]
         else:
             res = art_path
-        return path_or_none(res) if sign.type == Path else res
+        res = path_or_none(res) if sign.type == Path else res
     elif sign.type in [List[str], List[Path], Set[str], Set[Path]]:
         if path_object is None:
-            res = None
+            return None
         elif isinstance(path_object, str):
             res = [path_object]
         elif isinstance(path_object, list) and all([
@@ -96,17 +99,25 @@ def handle_input_artifact(name, sign, slices=None, data_root="/tmp",
             res = list(flatten(path_object).values())
 
         if sign.type == List[str]:
-            return res
+            pass
         elif sign.type == List[Path]:
-            return path_or_none(res)
+            res = path_or_none(res)
         elif sign.type == Set[str]:
-            return set(res)
+            res = set(res)
         else:
-            return set(path_or_none(res))
+            res = set(path_or_none(res))
     elif sign.type in [Dict[str, str], NestedDict[str]]:
-        return path_object
+        res = path_object
     elif sign.type in [Dict[str, Path], NestedDict[Path]]:
-        return path_or_none(path_object)
+        res = path_or_none(path_object)
+
+    if res is None:
+        return None
+
+    _cls = res.__class__
+    res = artifact_classes[_cls](res)
+    res.art_root = root
+    return res
 
 
 def path_or_none(p):
