@@ -5,7 +5,8 @@ from .op_template import OPTemplate
 from .step import Step
 
 try:
-    from argo.workflows.client import V1alpha1Arguments, V1alpha1DAGTask
+    from argo.workflows.client import (V1alpha1Arguments, V1alpha1ContinueOn,
+                                       V1alpha1DAGTask)
 except Exception:
     pass
 
@@ -51,12 +52,20 @@ class Task(Step):
         if self.prepare_step is not None:
             self.dependencies.append(self.prepare_step)
         if self.check_step is not None:
-            self.check_step.dependencies.append(self)
+            self.check_step.dependencies = [
+                "%s.Succeeded || %s.Failed || %s.Errored" % (self, self, self)]
 
     @classmethod
     def from_dict(cls, d, templates):
         task = super().from_dict(d, templates)
         task.dependencies = d.get("dependencies", [])
+        if d.get("depends"):
+            for dep in d["depends"].split("&&"):
+                # removeprefix and removesuffix only supported for python>=3.9
+                dep = dep[1:] if dep.startswith("(") else dep
+                dep = dep[:-1] if dep.endswith(")") else dep
+                dep = dep[:-10] if dep.endswith(".Succeeded") else dep
+                task.dependencies.append(dep)
         return task
 
     def set_parameters(self, parameters):
@@ -88,15 +97,29 @@ class Task(Step):
         depends = []
         for task in self.dependencies:
             if isinstance(task, Task):
-                depends.append("(%s.Succeeded)" % task)
+                if task.check_step is not None:
+                    depends.append("(%s.Succeeded)" % task.check_step)
+                else:
+                    depends.append("(%s.Succeeded)" % task)
             else:
                 depends.append("(%s)" % task)
+        if self.continue_on_failed or self.continue_on_error:
+            continue_on = V1alpha1ContinueOn(
+                failed=self.continue_on_failed, error=self.continue_on_error)
+        else:
+            continue_on = None
         return V1alpha1DAGTask(
             name=self.name, template=self.template.name,
             arguments=V1alpha1Arguments(
                 parameters=self.argo_parameters,
                 artifacts=self.argo_artifacts
             ), when=self.when, with_param=self.with_param,
+            continue_on=continue_on,
             with_sequence=self.with_sequence,
             depends=" && ".join(depends),
         )
+
+    def convert_to_graph(self):
+        g = super().convert_to_graph()
+        g["dependencies"] = [str(d) for d in self.dependencies]
+        return g
