@@ -17,7 +17,7 @@ import tempfile
 import uuid
 from abc import ABC
 from functools import partial
-from pathlib import Path
+from pathlib import Path, PosixPath, WindowsPath
 from typing import Dict, List, Optional, Set, Tuple, Union
 
 import jsonpickle
@@ -68,7 +68,6 @@ def download_artifact(
         sub_path: Optional[str] = None,
         slice: Optional[int] = None,
         path: os.PathLike = ".",
-        debug_download: bool = False,
         remove_catalog: bool = True,
         **kwargs,
 ) -> List[str]:
@@ -88,7 +87,7 @@ def download_artifact(
         bucket_name: bucket name for Minio
         skip_exists: skip files with the same MD5
     """
-    if config["mode"] == "debug" and not debug_download:
+    if getattr(artifact, "local_path", None) is not None:
         if config["debug_copy_method"] == "symlink":
             linktree(artifact.local_path, path)
         elif config["debug_copy_method"] == "link":
@@ -104,6 +103,7 @@ def download_artifact(
 
     if sub_path is not None:
         key = key + "/" + sub_path
+        # subpath may be a file
         path = os.path.join(path, os.path.dirname(sub_path))
         download_s3(key=key, recursive=True, path=path, keep_dir=True,
                     **kwargs)
@@ -239,12 +239,19 @@ def upload_artifact(
                     relpath = abspath[abspath.find(":")+2:]
             target = os.path.join(tmpdir, relpath)
             os.makedirs(os.path.dirname(target), exist_ok=True)
-            if config["mode"] == "debug" and cwd.startswith(p):
+            if config["mode"] == "debug" and not config["debug_s3"] and \
+                    cwd.startswith(os.path.abspath(str(p))):
                 # To avoid recursive symlink
                 copy_file(abspath, target)
             else:
-                if not (os.path.exists(target) and
-                        os.path.samefile(abspath, target)):
+                if os.path.exists(target) and os.path.samefile(
+                        abspath, target):
+                    pass
+                elif os.path.isdir(target):
+                    # there is overlap between pathes, keep parent only
+                    shutil.rmtree(target)
+                    os.symlink(abspath, target)
+                else:
                     os.symlink(abspath, target)
             path_list.append({"dflow_list_item": relpath.replace("\\", "/"),
                               "order": i})
@@ -254,12 +261,11 @@ def upload_artifact(
         with open(os.path.join(catalog_dir, str(uuid.uuid4())), "w") as f:
             f.write(jsonpickle.dumps({"path_list": path_list}))
 
-        if config["mode"] == "debug":
-            os.makedirs("upload", exist_ok=True)
-            resdir = shutil.move(tmpdir, "upload")
+        if config["mode"] == "debug" and not config["debug_s3"]:
+            path = upload_s3(tmpdir, debug_func=shutil.move)
             # To prevent exception in destruction
             os.makedirs(tmpdir, exist_ok=True)
-            return LocalArtifact(local_path=os.path.abspath(resdir))
+            return LocalArtifact(local_path=path)
 
         if archive == "tar":
             os.chdir(os.path.dirname(tmpdir))
@@ -281,7 +287,7 @@ def upload_artifact(
             urn = config["lineage"].register_artifact(
                 namespace, dataset_name, key, **kwargs)
         else:
-            logging.warn("Lineage client not provided")
+            logging.warning("Lineage client not provided")
 
     return S3Artifact(key=key, path_list=path_list, urn=urn)
 
@@ -371,8 +377,18 @@ def upload_s3(
         path: os.PathLike,
         key: Optional[str] = None,
         prefix: Optional[str] = None,
+        debug_func=os.symlink,
         **kwargs,
 ) -> str:
+    if config["mode"] == "debug" and not config["debug_s3"]:
+        if key is None:
+            key = "upload/%s/%s" % (uuid.uuid4(), os.path.basename(path))
+        target = os.path.abspath(os.path.join(
+            config["debug_workdir"], config["debug_artifact_dir"], key))
+        os.makedirs(os.path.dirname(target), exist_ok=True)
+        debug_func(os.path.abspath(path), target)
+        return target
+
     if s3_config["storage_client"] is not None:
         client = s3_config["storage_client"]
     else:
@@ -401,6 +417,8 @@ def upload_s3(
             for f in fs:
                 client.upload(key="%s%s/%s" % (key, rel_path, f),
                               path=os.path.join(dn, f))
+    else:
+        raise FileNotFoundError("No such file or directory: %s" % path)
     return key
 
 
@@ -523,7 +541,7 @@ def copy_file(src, dst, func=try_link):
     elif os.path.isfile(src):
         func(src, dst)
     else:
-        raise RuntimeError("File %s not found" % src)
+        raise FileNotFoundError("File %s not found" % src)
 
 
 def catalog_of_local_artifact(art_path, remove=False):
@@ -573,8 +591,8 @@ def remove_empty_dir_tag(path):
 
 
 def randstr(length: int = 5) -> str:
-    return "".join(random.sample(string.digits + string.ascii_lowercase,
-                                 length))
+    return "".join(random.choices(string.digits + string.ascii_lowercase,
+                                  k=length))
 
 
 @contextlib.contextmanager
@@ -653,7 +671,7 @@ def run_command(
         cmd = [str(x) for x in cmd]
 
     if try_bash:
-        arg = "-ic" if interactive else "-c"
+        arg = "-lc" if interactive else "-c"
         script = "if command -v bash 2>&1 >/dev/null; then bash %s " % arg + \
             shlex.quote(" ".join(cmd)) + "; else " + " ".join(cmd) + "; fi"
         cmd = [script]
@@ -736,9 +754,10 @@ def linktree(src, dst, func=os.symlink):
 
 
 def force_link(src, dst, func=os.symlink):
-    if os.path.exists(dst):
-        if os.path.samefile(src, dst):
-            return
+    # avoid cross link
+    if os.path.exists(dst) and os.path.samefile(src, dst):
+        return
+    if os.path.islink(dst):
         os.remove(dst)
     func(src, dst)
 
@@ -805,3 +824,64 @@ class MinioClient(StorageClient):
     def get_md5(self, key: str) -> str:
         return self.client.stat_object(bucket_name=self.bucket_name,
                                        object_name=key).etag
+
+
+class ArtifactStr(str):
+    pass
+
+
+class ArtifactPosixPath(PosixPath):
+    pass
+
+
+class ArtifactWindowsPath(WindowsPath):
+    pass
+
+
+class ArtifactList(list):
+    pass
+
+
+class ArtifactSet(set):
+    pass
+
+
+class ArtifactDict(dict):
+    pass
+
+
+artifact_classes = {
+    str: ArtifactStr,
+    PosixPath: ArtifactPosixPath,
+    WindowsPath: ArtifactWindowsPath,
+    list: ArtifactList,
+    set: ArtifactSet,
+    dict: ArtifactDict,
+}
+
+
+class Variable:
+    def __init__(self, expr):
+        self.expr = expr
+
+    def evalable_repr(self, imports):
+        return self.expr
+
+
+def evalable_repr(obj, imports):
+    if obj is None or isinstance(obj, (str, int, float, bool)):
+        return repr(obj)
+    if isinstance(obj, list):
+        return "[%s]" % ", ".join([evalable_repr(i, imports) for i in obj])
+    if isinstance(obj, set):
+        if len(obj) == 0:
+            return "set()"
+        else:
+            return "{%s}" % ", ".join([evalable_repr(i, imports) for i in obj])
+    if isinstance(obj, dict):
+        return "{%s}" % ", ".join(["%s: %s" % (evalable_repr(
+            k, imports), evalable_repr(v, imports)) for k, v in obj.items()])
+    if hasattr(obj, "evalable_repr"):
+        return obj.evalable_repr(imports)
+    imports.add((None, "jsonpickle"))
+    return "jsonpickle.loads(%s)" % repr(jsonpickle.dumps(obj))
