@@ -1,7 +1,6 @@
 import json
 import logging
 import os
-import signal
 import sys
 import time
 from copy import deepcopy
@@ -125,6 +124,8 @@ class Workflow:
         if annotations is None:
             annotations = deepcopy(config["workflow_annotations"])
         self.annotations = annotations
+        if labels is None:
+            labels = {}
         self.labels = labels
         self.parallelism = parallelism
         self.pod_gc_strategy = pod_gc_strategy
@@ -219,12 +220,35 @@ class Workflow:
             if self.id is None:
                 while True:
                     self.id = self.name + "-" + randstr()
-                    wfdir = os.path.abspath(self.id)
+                    wfdir = os.path.abspath(os.path.join(
+                        config["debug_workdir"], self.id))
                     if not os.path.exists(wfdir):
                         os.makedirs(wfdir)
                         break
             else:
-                wfdir = os.path.abspath(self.id)
+                wfdir = os.path.abspath(os.path.join(
+                    config["debug_workdir"], self.id))
+                if os.path.exists(wfdir):
+                    with open(os.path.join(wfdir, "status"), "r") as f:
+                        status = f.read()
+                    if status == "Succeeded":
+                        logger.warning(
+                            "Workflow %s has been succeeded" % self.id)
+                        return
+                    with open(os.path.join(wfdir, "pid"), "r") as f:
+                        pid = int(f.read())
+                    import psutil
+                    try:
+                        p = psutil.Process(pid)
+                        ps = p.status()
+                    except psutil.NoSuchProcess:
+                        ps = None
+                    logger.warning("Workflow %s process %s is %s" % (
+                        self.id, pid, ps))
+                    if ps == psutil.STATUS_RUNNING:
+                        logger.warning("Do nothing")
+                        return
+                    logger.warning("Restart workflow %s" % self.id)
                 os.makedirs(wfdir, exist_ok=True)
 
             if reuse_step is not None:
@@ -260,7 +284,9 @@ class Workflow:
                         if "dflow_group_key" in step.inputs.parameters:
                             key = step.inputs.parameters[
                                 "dflow_group_key"].value
-                            if not os.path.exists(os.path.join(wfdir, key)):
+                            if not os.path.exists(os.path.join(wfdir, key)) \
+                                and not os.path.islink(os.path.join(
+                                    wfdir, key)):
                                 os.symlink(
                                     os.path.join(
                                         os.path.abspath(step.workflow), key),
@@ -336,6 +362,8 @@ class Workflow:
         self.uid = workflow.metadata.uid
         print("Workflow has been submitted (ID: %s, UID: %s)" % (self.id,
                                                                  self.uid))
+        print("Workflow link: %s/workflows/%s/%s" % (self.host, self.namespace,
+                                                     self.id))
         return workflow
 
     def wait(self, interval=1):
@@ -472,8 +500,9 @@ class Workflow:
 
             self.handle_template(self.entrypoint, memoize_prefix=self.id,
                                  memoize_configmap="dflow")
-            status = {"outputs": {"parameters": [
-                {"name": key, "value": id} for key, id in key2id.items()]}}
+            if config["save_keys_in_global_outputs"]:
+                status = {"outputs": {"parameters": [
+                    {"name": key, "value": id} for key, id in key2id.items()]}}
         else:
             self.handle_template(self.entrypoint)
 
@@ -514,7 +543,7 @@ class Workflow:
                         'POST', body=secret, response_type='V1Secret',
                         header_params=config["http_headers"],
                         _return_http_data_only=True)
-                    self.image_pull_secrets[i] = s.name
+                    self.image_pull_secrets[i] = V1LocalObjectReference(s.name)
 
         if config["register_tasks"]:
             workflow_urn = config["lineage"].register_workflow(self.name)
@@ -535,8 +564,7 @@ class Workflow:
                 parallelism=self.parallelism,
                 volume_claim_templates=argo_pvcs,
                 pod_gc=V1alpha1PodGC(strategy=self.pod_gc_strategy),
-                image_pull_secrets=None if self.image_pull_secrets is None else
-                [V1LocalObjectReference(s) for s in self.image_pull_secrets],
+                image_pull_secrets=self.image_pull_secrets,
                 artifact_repository_ref=None if self.artifact_repo_key is None
                 else V1alpha1ArtifactRepositoryRef(key=self.artifact_repo_key)
             ),
@@ -711,6 +739,94 @@ class Workflow:
                     if pvc.name not in self.pvcs:
                         self.pvcs[pvc.name] = pvc
 
+    def get_graph_templates(self, template, graph_templates=None):
+        if graph_templates is None:
+            graph_templates = {}
+        if template.name not in graph_templates:
+            if isinstance(template, (Steps, DAG)):
+                graph_template, templates = template.convert_to_graph()
+                graph_templates[template.name] = graph_template
+                for t in templates:
+                    self.get_graph_templates(t, graph_templates)
+            else:
+                graph_templates[template.name] = template.convert_to_graph()
+        return graph_templates
+
+    def to_graph(self):
+        graph_templates = self.get_graph_templates(self.entrypoint)
+        g = {
+            "name": self.name,
+            "namespace": self.namespace,
+            "id": self.id,
+            "context": self.context,
+            "annotations": self.annotations,
+            "labels": self.labels,
+            "parallelism": self.parallelism,
+            "pod_gc_strategy": self.pod_gc_strategy,
+            "artifact_repo_key": self.artifact_repo_key,
+            "image_pull_secrets": self.image_pull_secrets,
+            "parameters": self.parameters,
+            "entrypoint": self.entrypoint.name,
+            "templates": graph_templates,
+        }
+        return json.loads(jsonpickle.dumps(g, make_refs=False))
+
+    def to_graph_json(self, **kwargs):
+        return json.dumps(self.to_graph(), **kwargs)
+
+    def to_graph_yaml(self, **kwargs):
+        return yaml.dump(self.to_graph(), **kwargs)
+
+    @classmethod
+    def from_graph(cls, graph):
+        from .python import PythonOPTemplate
+
+        graph = jsonpickle.loads(json.dumps(graph))
+        templates = {}
+        for name, template in list(graph["templates"].items()):
+            if template["type"] == "PythonOPTemplate":
+                templates[name] = PythonOPTemplate.from_graph(template)
+                del graph["templates"][name]
+            elif template["type"] == "ScriptOPTemplate":
+                templates[name] = ScriptOPTemplate.from_graph(template)
+                del graph["templates"][name]
+
+        while len(graph["templates"]) > 0:
+            update = False
+            for name, template in list(graph["templates"].items()):
+                if template["type"] == "Steps":
+                    if all([all([
+                        ps["template"] in templates or ps["template"] == name
+                            for ps in s]) for s in template["steps"]]):
+                        templates[name] = Steps.from_graph(template, templates)
+                        del graph["templates"][name]
+                        update = True
+                elif template["type"] == "DAG":
+                    if all([
+                        t["template"] in templates or t["template"] == name
+                            for t in template["tasks"]]):
+                        templates[name] = DAG.from_graph(template, templates)
+                        del graph["templates"][name]
+                        update = True
+            assert update, "Failed to resolve templates: %s" % list(
+                graph["templates"])
+
+        del graph["templates"]
+        entrypoint = templates[graph.pop("entrypoint")]
+        if isinstance(entrypoint, Steps):
+            graph["steps"] = entrypoint
+        elif isinstance(entrypoint, DAG):
+            graph["dag"] = entrypoint
+        return cls(**graph)
+
+    @classmethod
+    def from_graph_json(cls, j, **kwargs):
+        return cls.from_graph(json.loads(j, **kwargs))
+
+    @classmethod
+    def from_graph_yaml(cls, y, **kwargs):
+        return cls.from_graph(yaml.full_load(y, **kwargs))
+
     def query(
             self,
             fields: Optional[List[str]] = None,
@@ -754,7 +870,8 @@ class Workflow:
             Pending, Running, Succeeded, Failed, Error, etc
         """
         if config["mode"] == "debug":
-            with open("%s/status" % self.id, "r") as f:
+            wfdir = os.path.join(config["debug_workdir"], self.id)
+            with open(os.path.join(wfdir, "status"), "r") as f:
                 return f.read()
         workflow = self.query(fields=["status.phase"])
 
@@ -791,11 +908,12 @@ class Workflow:
             a list of steps
         """
         if config["mode"] == "debug":
+            wfdir = os.path.join(config["debug_workdir"], self.id)
             if key is not None and not isinstance(key, list):
                 key = [key]
             step_list = []
-            for s in os.listdir(self.id):
-                stepdir = os.path.join(self.id, s)
+            for s in os.listdir(wfdir):
+                stepdir = os.path.join(wfdir, s)
                 if not os.path.isdir(stepdir):
                     continue
                 if not os.path.exists(os.path.join(stepdir, "name")):
@@ -851,7 +969,8 @@ class Workflow:
                                         stepdir, io, "parameters/.dflow", p),
                                         "r") as f:
                                     _type = json.load(f)["type"]
-                                if _type != str(str):
+                                # for backward compatible
+                                if _type not in ["str", str(str)]:
                                     val = jsonpickle.loads(val)
                             step[io]["parameters"].append({
                                 "name": p, "value": val, "type": _type})
@@ -899,8 +1018,8 @@ class Workflow:
             return [par["name"] for par in
                     response["status"]["outputs"]["parameters"]]
         except Exception:
-            logger.warn("Key-ID map not found in the global outputs, downgrade"
-                        " to full query")
+            logger.warning("Key-ID map not found in the global outputs, "
+                           "downgrade to full query")
             return [step.key for step in self.query_step()
                     if step.key is not None]
 
@@ -939,8 +1058,8 @@ class Workflow:
                     'status.nodes.' + key2id[k] for k in key])
             return workflow.get_step(name=name, phase=phase, id=id, type=type)
         except Exception:
-            logger.warn("Key-ID map not found in the global outputs, downgrade"
-                        " to full query")
+            logger.warning("Key-ID map not found in the global outputs, "
+                           "downgrade to full query")
             return self.query_step(key=key, name=name, phase=phase, id=id,
                                    type=type)
 
@@ -954,6 +1073,35 @@ class Workflow:
         Returns:
             a list of steps
         """
+        if config["mode"] == "debug":
+            wfdir = os.path.join(config["debug_workdir"], self.id)
+            if not os.path.exists(os.path.join(wfdir, "outputs")):
+                return None
+            outputs = {"parameters": [], "artifacts": []}
+            pars = os.path.join(wfdir, "outputs", "parameters")
+            if os.path.exists(pars):
+                for p in os.listdir(pars):
+                    if p == ".dflow":
+                        continue
+                    with open(os.path.join(pars, p), "r") as f:
+                        val = f.read()
+                    _type = None
+                    if os.path.exists(os.path.join(pars, ".dflow", p)):
+                        with open(os.path.join(pars, ".dflow", p), "r") as f:
+                            _type = json.load(f)["type"]
+                        # for backward compatible
+                        if _type != ["str", str(str)]:
+                            val = jsonpickle.loads(val)
+                    outputs["parameters"].append({
+                        "name": p, "value": val, "type": _type})
+            arts = os.path.join(wfdir, "outputs", "artifacts")
+            if os.path.exists(arts):
+                for a in os.listdir(arts):
+                    outputs["artifacts"].append({
+                        "name": a,
+                        "local_path": os.path.abspath(os.path.join(arts, a))})
+            step = ArgoStep({"outputs": outputs}, self.id)
+            return step.outputs
         workflow = self.query(fields=['metadata.name', 'status.outputs'])
         step = ArgoStep(workflow.status, workflow.metadata.name)
         if hasattr(step, "outputs"):
@@ -968,9 +1116,14 @@ class Workflow:
         if self.id is None:
             raise RuntimeError("Workflow ID is None")
         if config["mode"] == "debug":
-            with open(os.path.join(self.id, "pid"), "r") as f:
+            wfdir = os.path.join(config["debug_workdir"], self.id)
+            with open(os.path.join(wfdir, "pid"), "r") as f:
                 pid = int(f.read())
-            os.killpg(os.getpgid(pid), signal.SIGTERM)
+            import psutil
+            p = psutil.Process(pid)
+            for c in p.children(recursive=True):
+                c.terminate()
+            p.terminate()
             return
         self.api_instance.api_client.call_api(
             '/api/v1/workflows/%s/%s/terminate' % (self.namespace, self.id),
